## Uses the huggingface text generation inference API
import os, json
from enum import Enum
import requests
from litellm import logging
import time 
from typing import Callable
from litellm.utils import ModelResponse

class HuggingfaceError(Exception):
    def __init__(self, status_code, message):
        self.status_code = status_code
        self.message = message
        super().__init__(self.message) # Call the base class constructor with the parameters it needs

class HuggingfaceRestAPILLM():
    def __init__(self, encoding, api_key=None) -> None:
        self.encoding = encoding
        self.validate_environment(api_key=api_key)

    def validate_environment(self, api_key): # set up the environment required to run the model 
        self.headers = {
            "content-type": "application/json",
        }
        # get the api key if it exists in the environment or is passed in, but don't require it
        self.api_key = api_key
        if self.api_key != None:
            self.headers["Authorization"] = f"Bearer {self.api_key}" 

    def completion(self, model: str, messages: list, custom_api_base: str, model_response: ModelResponse, print_verbose: Callable, optional_params=None, litellm_params=None, logger_fn=None): # logic for parsing in - calling - parsing out model completion calls
        if custom_api_base:
            completion_url = custom_api_base
        elif "HF_API_BASE" in os.environ:
            completion_url = os.getenv("HF_API_BASE")
        else:
            completion_url = f"https://api-inference.huggingface.co/models/{model}"
        prompt = ""
        if "meta-llama" in model and "chat" in model: # use the required special tokens for meta-llama - https://huggingface.co/blog/llama2#how-to-prompt-llama-2
            prompt = "<s>"
            for message in messages:
                if message["role"] == "system":
                    prompt += "[INST] <<SYS>>" + message["content"]
                elif message["role"] == "assistant":
                    prompt += message["content"] + "</s><s>[INST]"
                elif message["role"] == "user":
                    prompt += message["content"] + "[/INST]"
        else:
            for message in messages:
                prompt += f"{message['content']}"
        ### MAP INPUT PARAMS 
        # max tokens  
        if "max_tokens" in optional_params:
            value = optional_params.pop("max_tokens")
            optional_params["max_new_tokens"] = value
        data = {
            "inputs": prompt,
            # "parameters": optional_params
        }
        ## LOGGING
        logging(model=model, input=prompt, additional_args={"litellm_params": litellm_params, "optional_params": optional_params}, logger_fn=logger_fn)
        ## COMPLETION CALL
        response = requests.post(completion_url, headers=self.headers, data=json.dumps(data))
        if "stream" in optional_params and optional_params["stream"] == True:
            return response.iter_lines()
        else:
            ## LOGGING
            logging(model=model, input=prompt, additional_args={"litellm_params": litellm_params, "optional_params": optional_params, "original_response": response.text}, logger_fn=logger_fn)
            print_verbose(f"raw model_response: {response.text}")
            ## RESPONSE OBJECT
            completion_response = response.json()
            print_verbose(f"response: {completion_response}")
            if isinstance(completion_response, dict) and "error" in completion_response:
                print_verbose(f"completion error: {completion_response['error']}")
                print_verbose(f"response.status_code: {response.status_code}")
                raise HuggingfaceError(message=completion_response["error"], status_code=response.status_code)
            else:
                model_response["choices"][0]["message"]["content"] = completion_response[0]["generated_text"]    
            
            ## CALCULATING USAGE
            prompt_tokens = len(self.encoding.encode(prompt)) ##[TODO] use the llama2 tokenizer here
            completion_tokens = len(self.encoding.encode(model_response["choices"][0]["message"]["content"])) ##[TODO] use the llama2 tokenizer here
            
            
            model_response["created"] = time.time()
            model_response["model"] = model
            model_response["usage"] = {
                "prompt_tokens": prompt_tokens,
                "completion_tokens": completion_tokens,
                "total_tokens": prompt_tokens + completion_tokens
                }
            return model_response
        pass

    def embedding(): # logic for parsing in - calling - parsing out model embedding calls
        pass 


import litellm
import time 
from concurrent.futures import ThreadPoolExecutor
import traceback

def testing_batch_completion(*args, **kwargs):
  try:
    batch_models = args[0] if len(args) > 0 else kwargs.pop("models") ## expected input format- ["gpt-3.5-turbo", {"model": "qvv0xeq", "custom_llm_provider"="baseten"}...]
    batch_messages = args[1] if len(args) > 1 else kwargs.pop("messages")
    results = []
    completions = []
    exceptions = []
    times = []
    with ThreadPoolExecutor() as executor:
        for model in batch_models:
            kwargs_modified = dict(kwargs)
            args_modified = list(args)
            if len(args) > 0:
                args_modified[0] = model["model"]
            else: 
                kwargs_modified["model"] = model["model"] if isinstance(model, dict) and "model" in model else model # if model is a dictionary get it's value else assume it's a string
                kwargs_modified["custom_llm_provider"] = model["custom_llm_provider"] if isinstance(model, dict) and "custom_llm_provider" in model else None
                kwargs_modified["custom_api_base"] = model["custom_api_base"] if isinstance(model, dict) and "custom_api_base" in model else None
            for message_list in batch_messages:
                if len(args) > 1:
                    args_modified[1] = message_list
                    future = executor.submit(litellm.completion, *args_modified, **kwargs_modified)
                else:
                    kwargs_modified["messages"] = message_list
                    future = executor.submit(litellm.completion, *args_modified, **kwargs_modified)
                completions.append((future, message_list))
    
    # Retrieve the results and calculate elapsed time for each completion call
    for completion in completions:
        future, message_list = completion
        start_time = time.time()
        try:
            result = future.result()
            end_time = time.time()
            elapsed_time = end_time - start_time
            result_dict = {"status": "succeeded", "response": future.result(), "prompt": message_list, "response_time": elapsed_time}
            results.append(result_dict)
        except Exception as e:
            end_time = time.time()
            elapsed_time = end_time - start_time
            result_dict = {"status": "failed", "response": e, "response_time": elapsed_time}
            results.append(result_dict)
    return results
  except:
     traceback.print_exc()

def duration_test_model(original_function):
    def wrapper_function(*args, **kwargs):
        # Code to be executed before the original function
        duration = kwargs.pop("duration", None)
        interval = kwargs.pop("interval", None)
        results = []
        if duration and interval:
            start_time = time.time()
            end_time = start_time + duration  # default to 1hr duration
            while time.time() < end_time:
                result = original_function(*args, **kwargs)
                results.append(result)
                time.sleep(interval)
        else:
            result = original_function(*args, **kwargs)
            results = result
        return results

    # Return the wrapper function
    return wrapper_function

@duration_test_model
def load_test_model(models: list, prompt: str = None, num_calls: int = None):
  test_calls = 100
  if num_calls:
     test_calls = num_calls
  input_prompt = prompt if prompt else "Hey, how's it going?"
  messages = [{"role": "user", "content": prompt}] if prompt else [{"role": "user", "content": input_prompt}]
  full_message_list = [messages for _ in range(test_calls)] # call it as many times as set by user to load test models
  start_time = time.time()
  try:
    results = testing_batch_completion(models=models, messages=full_message_list)
    end_time = time.time() 
    response_time = end_time - start_time
    return {"total_response_time": response_time, "calls_made": test_calls, "prompt": input_prompt, "results": results}
  except Exception as e:
    traceback.print_exc()
    end_time = time.time() 
    response_time = end_time - start_time
    return {"total_response_time": response_time, "calls_made": test_calls, "prompt": input_prompt, "exception": e}


## LiteLLM versions of the OpenAI Exception Types
from openai.error import AuthenticationError, InvalidRequestError, RateLimitError, ServiceUnavailableError, OpenAIError

class AuthenticationError(AuthenticationError):
    def __init__(self, message, llm_provider):
        self.status_code = 401
        self.message = message
        self.llm_provider = llm_provider
        super().__init__(self.message) # Call the base class constructor with the parameters it needs


class InvalidRequestError(InvalidRequestError):
    def __init__(self, message, model, llm_provider):
        self.status_code = 400
        self.message = message
        self.model = model
        self.llm_provider = llm_provider
        super().__init__(self.message, f"{self.model}") # Call the base class constructor with the parameters it needs


class RateLimitError(RateLimitError):
    def __init__(self, message, llm_provider):
        self.status_code = 429
        self.message = message
        self.llm_provider = llm_provider
        super().__init__(self.message) # Call the base class constructor with the parameters it needs

class ServiceUnavailableError(ServiceUnavailableError):
    def __init__(self, message, llm_provider):
        self.status_code = 500
        self.message = message
        self.llm_provider = llm_provider
        super().__init__(self.message) # Call the base class constructor with the parameters it needs

class OpenAIError(OpenAIError):
    def __init__(self, original_exception):
        self.status_code = original_exception.http_status
        super().__init__(http_body=original_exception.http_body,
                         http_status=original_exception.http_status,
                         json_body=original_exception.json_body,
                         headers=original_exception.headers,
                         code=original_exception.code)
        self.llm_provider = "openai"


import sys, os
import traceback
from dotenv import load_dotenv
load_dotenv()
import os
sys.path.insert(0, os.path.abspath('../..'))  # Adds the parent directory to the system path
import pytest
import litellm
from litellm import embedding, completion

litellm.caching = True
messages = [{"role": "user", "content": "who is ishaan Github?  "}]



# test if response cached
def test_caching():
    try:
        response1 = completion(model="gpt-3.5-turbo", messages=messages)
        response2 = completion(model="gpt-3.5-turbo", messages=messages)
        print(f"response1: {response1}")
        print(f"response2: {response2}")
        litellm.caching = False
        if response2 != response1:
            print(f"response1: {response1}")
            print(f"response2: {response2}")
            pytest.fail(f"Error occurred: {e}")
    except Exception as e:
        litellm.caching = False
        print(f"error occurred: {traceback.format_exc()}") 
        pytest.fail(f"Error occurred: {e}")




